//===-- libubsan.c -----------------------------------------------*- C -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// Entry points to the runtime library for Clang's undefined behavior sanitizer.
//
//===----------------------------------------------------------------------===//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SUMARY_UBSAN "Sanitizer UBSan Warning: "
#define ERROR_MEG_MAX 200

/**
 *  winAMS確認用グローバル変数 extern宣言
 */
extern char ubsan_error_message[ERROR_MEG_MAX];

void UbsanTriggerLog(){}
/**
 * ubsan handler
 * handler関数コメントはLLVM project から引用しています。
 * https://github.com/llvm/llvm-project/blob/main/compiler-rt/lib/ubsan/ubsan_handlers.h
 */

/**
 * clang ubsan 構造体定義
 * */
struct SourceLocation{
	const char *file_name;
	uint32_t line;
	uint32_t column;
};

struct TypeDescriptor {
	uint16_t type_kind;
	uint16_t type_info;
	char type_name[1];
};

struct TypeMismatchData {
	struct SourceLocation Loc;
	struct TypeDescriptor *type;
	unsigned int alignment;
	unsigned char type_check_kind;
};

struct OverflowData {
	struct SourceLocation Loc;
	const struct TypeDescriptor *Type;
};

struct ShiftOutOfBoundsData{
	struct SourceLocation Loc;
	const struct TypeDescriptor *LHSType;
	const struct TypeDescriptor *RHSType;
};

struct OutOfBoundsData{
	struct SourceLocation Loc;
	const struct TypeDescriptor *ArrayType;
	const struct TypeDescriptor *IndexType;	
};

struct PointerOverflowData{
	struct SourceLocation Loc;
};

typedef unsigned long long uptr;
typedef signed long long sptr;
typedef uptr ValueHandle;

/*
 * When compiling with -fsanitize=undefined the compiler expects functions
 * with the following signatures. The functions are never called directly,
 * only when undefined behavior is detected in instrumented code.
 */
void __ubsan_handle_type_mismatch_v1(struct TypeMismatchData *Data, ValueHandle Pointer);
void __ubsan_handle_add_overflow(struct OverflowData *Data, ValueHandle LHS, ValueHandle RHS);
void __ubsan_handle_sub_overflow(struct OverflowData *Data, ValueHandle LHS, ValueHandle RHS);
void __ubsan_handle_mul_overflow(struct OverflowData *Data, ValueHandle LHS, ValueHandle RHS);
void __ubsan_handle_shift_out_of_bounds(struct ShiftOutOfBoundsData *Data, ValueHandle LHS, ValueHandle RHS);
void __ubsan_handle_out_of_bounds(struct OutOfBoundsData *Data, ValueHandle Index);
void __ubsan_handle_pointer_overflow(struct PointerOverflowData *Data, ValueHandle Base, ValueHandle Result);


/**
 *  function:__ubsan_handle_type_mismatch_v1
 *  \brief Handle a runtime type check failure, caused by either a misaligned pointer,
 *   a null pointer, or a pointer to insufficient storage for the type.
 * 
 * 　ランタイム型チェックの失敗を処理します。ポインタの位置がずれている、ポインタがnullである、
 *   または型に必要なメモリが不足していることによって引き起こされます。
 */
void __ubsan_handle_type_mismatch_v1(struct TypeMismatchData *Data,
									ValueHandle Pointer)
{	
	char *_ubsan_message_template = SUMARY_UBSAN "type mismatch data had occurred. [%s]"
									"%s:%d:%d ";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);
	//for WinAMS Macro log
	UbsanTriggerLog();
}

/**
 * function:__ubsan_handle_add_overflow
 * brief Handle an integer addition overflow.
 * 整数の加算オーバーフローを処理します
 */
void __ubsan_handle_add_overflow(struct OverflowData *Data,
								ValueHandle LHS, ValueHandle RHS)
{
	char *_ubsan_message_template = SUMARY_UBSAN "unsigned integer(add +) overflow had occurred.[%s]"
									"%s:%d:%d ";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);
	//for WinAMS Macro log
	UbsanTriggerLog();
}

/**
 * function:__ubsan_handle_sub_overflow
 * brief Handle an integer subtraction overflow.
 * 整数の減算オーバーフローを処理します
 */
void __ubsan_handle_sub_overflow(struct OverflowData *Data,
								ValueHandle LHS, ValueHandle RHS)
{
	char *_ubsan_message_template = SUMARY_UBSAN "unsigned integer(sub -) overflow had occurred.[%s]"
									"%s:%d:%d ";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);
	//for WinAMS Macro log
	UbsanTriggerLog();
}

/**
 * function:__ubsan_handle_mul_overflow
 * brief Handle an integer multiplication overflow.
 * 整数の乗算オーバーフローを処理します
 */
void __ubsan_handle_mul_overflow(struct OverflowData *Data,
								ValueHandle LHS, ValueHandle RHS)
{
	char *_ubsan_message_template = SUMARY_UBSAN "unsigned integer(mul *) overflow had occurred.[%s]"
									"%s:%d:%d";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);
	//for WinAMS Macro log
	UbsanTriggerLog();

}

/**
 * function:__ubsan_handle_shift_out_of_bounds
 * brief Handle a shift where the RHS is out of bounds or a left shift where the LHS is negative or overflows.
 * RHSが範囲外のシフト、またはLHSが負またはオーバーフローの左シフトを扱う
*/
void __ubsan_handle_shift_out_of_bounds(struct ShiftOutOfBoundsData *Data,
										 ValueHandle LHS,
										 ValueHandle RHS)
{

	char *_ubsan_message_template = SUMARY_UBSAN 
									"shift out of bounds had occurred.[%s] %s:%d:%d";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);

	//for WinAMS Macro log
	UbsanTriggerLog();
}

/**
 * function:__ubsan_handle_out_of_bounds
 * brief Handle an array index out of bounds error.
 * 配列インデックスの境界外エラーを処理
*/
void __ubsan_handle_out_of_bounds(struct OutOfBoundsData *Data,
									ValueHandle Index)
{

	char *_ubsan_message_template = SUMARY_UBSAN
									"array index out of bounds had occurred.[%s] %s:%d:%d";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);

	//for WinAMS Macro log
	UbsanTriggerLog();
}

/**
 * function:__ubsan_handle_pointer_overflow
 * ポインターオーバーフローを処理
*/
void __ubsan_handle_pointer_overflow(struct PointerOverflowData *Data,
									 ValueHandle Base,
									 ValueHandle Result)
{

	char *_ubsan_message_template = "%s:%d:%d " SUMARY_UBSAN 
									"pointer overflow had occurred.[%s]";

	memset(ubsan_error_message,'\0',sizeof(ubsan_error_message));

	//メッセージ書き込み		
	sprintf(
		ubsan_error_message, 				//WinAMS出力用メッセージ
		_ubsan_message_template,			//WinAMS出力用メッセージテンプレート
		__func__,			 				//検出ハンドラ関数名
		Data->Loc.file_name,      			//検出ファイル名
		Data->Loc.line,						//検出行数
		Data->Loc.column		 			//検出桁数
		);

	//for WinAMS Macro log
	UbsanTriggerLog();
}